@@ -54,6 +54,23 @@ use constant PFORDER_ACCEPT      => 10;
 
 ###########################################################################
 
+# change to 1 to enable the below test instrumentation points
+use constant SUPPORT_TEST_INSTRUMENTATION => 0;
+
+# test instrumentation point: simulate random child failures in 1 in
+# every N lookups
+our $TEST_MODE_CAUSE_RANDOM_KID_FAILURES = 0;
+
+# test instrumentation point: simulate child->parent and parent->child
+# write failures (needing retries) once in every N syswrite()s
+our $TEST_MODE_CAUSE_RANDOM_WRITE_RETRIES = 0;
+
+# test instrumentation point: simulate ping failures (for unspecified
+# reasons) once in every N pings
+our $TEST_MODE_CAUSE_RANDOM_PING_FAILURES = 0;
+
+###########################################################################
+
 # we use the following protocol between the master and child processes to
 # control when they accept/who accepts: server tells a child to accept with a
 # PF_ACCEPT_ORDER, child responds with "B$pid\n" when it's busy, and "I$pid\n"
@@ -109,6 +126,7 @@ sub add_child {
 sub child_exited {
   my ($self, $pid) = @_;
 
+  dbg("prefork: child $pid: just exited");
   delete $self->{kids}->{$pid};
 
   # note this for the select()-caller's benefit
@@ -144,6 +162,11 @@ sub child_error_kill {
     or warn "prefork: kill of failed child $pid failed: $!\n";
 
   $self->{backchannel}->delete_socket_for_child($pid);
+
+  if (defined $sock && defined $sock->fileno()) {
+    $self->{backchannel}->remove_from_selector($sock);
+  }
+
   if ($sock) {
     $sock->close;
   }
@@ -187,9 +210,16 @@ sub compute_lowest_child_pid {
 ###########################################################################
 
 sub set_server_fh {
-  my ($self, $fh) = @_;
-  $self->{server_fh} = $fh;
-  $self->{server_fileno} = $fh->fileno();
+  my ($self, @fhs) = @_;
+
+  $self->{server_fh} = [];
+  $self->{server_fileno} = [];
+
+  foreach my $fh (@fhs) {
+    next unless defined $fh;
+    push @{$self->{server_fh}}, $fh;
+    push @{$self->{server_fileno}}, $fh->fileno();
+  }
 }
 
 sub main_server_poll {
@@ -199,7 +229,7 @@ sub main_server_poll {
   if ($self->{overloaded}) {
     # don't select on the server fh -- we already KNOW that's ready,
     # since we're overloaded
-    vec($rin, $self->{server_fileno}, 1) = 0;
+    $self->vec_all(\$rin, $self->{server_fileno}, 0);
   }
 
   my ($rout, $eout, $nfound, $timeleft, $selerr);
@@ -212,7 +242,8 @@ sub main_server_poll {
 
   $timer->run(sub {
 
-    $self->{child_just_exited} = 0;
+    # right before select() syscall, but after alarm(), eval scope, etc.
+    $self->{child_just_exited} = 0;     
     ($nfound, $timeleft) = select($rout=$rin, undef, $eout=$rin, $tout);
     $selerr = $!;
 
@@ -265,7 +296,7 @@ sub main_server_poll {
 
   # errors on the handle?
   # return them immediately, they may be from a SIGHUP restart signal
-  if (vec ($eout, $self->{server_fileno}, 1)) {
+  if ($self->vec_all(\$eout, $self->{server_fileno})) {
     warn "prefork: select returned error on server filehandle: $selerr $!\n";
     return;
   }
@@ -283,7 +314,7 @@ sub main_server_poll {
   }
 
   # were the kids ready, or did we get signal?
-  if (vec ($rout, $self->{server_fileno}, 1)) {
+  if ($self->vec_all(\$rout, $self->{server_fileno})) {
     # dbg("prefork: server fh ready");
     # the server socket: new connection from a client
     if (!$self->order_idle_child_to_accept()) {
@@ -333,15 +364,23 @@ sub main_ping_kids {
 
   $self->{server_last_ping} = $now;
 
+  keys %{$self->{backchannel}->{kids}};     # reset each() iterator
   my ($sock, $kid);
   while (($kid, $sock) = each %{$self->{backchannel}->{kids}}) {
     # if the file handle is still defined ping the child
     # bug 4852: if not, we've run into a race condition with the child's
     # SIGCHLD handler... try killing again just in case something else happened
-    if (defined $sock && defined $sock->fileno) {
+
+    if (SUPPORT_TEST_INSTRUMENTATION && $TEST_MODE_CAUSE_RANDOM_PING_FAILURES &&
+              rand $TEST_MODE_CAUSE_RANDOM_PING_FAILURES < 1)
+    {
+      warn "prefork: TEST_MODE_CAUSE_RANDOM_PING_FAILURES simulating ping failure";
+    }
+    elsif (defined $sock && defined $sock->fileno) {
       $self->syswrite_with_retry($sock, PF_PING_ORDER, $kid, 3) and next;
       warn "prefork: write of ping failed to $kid fd=".$sock->fileno.": ".$!;
-    } else {
+    }
+    else {
       warn "prefork: cannot ping $kid, file handle not defined, child likely ".
 	   "to still be processing SIGCHLD handler after killing itself\n";
     }
@@ -365,7 +404,7 @@ sub read_one_message_from_child_socket {
     # stop it being select'd
     my $fno = $sock->fileno;
     if (defined $fno) {
-      vec(${$self->{backchannel}->{selector}}, $fno, 1) = 0;
+      $self->{backchannel}->remove_from_selector($sock);
       $sock->close();
     }
 
@@ -402,6 +441,13 @@ sub order_idle_child_to_accept {
   if (defined $kid)
   {
     my $sock = $self->{backchannel}->get_socket_for_child($kid);
+
+    if (SUPPORT_TEST_INSTRUMENTATION && $TEST_MODE_CAUSE_RANDOM_KID_FAILURES) {
+      if (rand $TEST_MODE_CAUSE_RANDOM_KID_FAILURES < 1) {
+        $sock = undef; warn "prefork: TEST_MODE_CAUSE_RANDOM_KID_FAILURES simulating no socket for kid $kid";
+      }
+    }
+
     if (!$sock)
     {
       # this should not happen, but if it does, trap it here
@@ -491,8 +537,6 @@ sub update_child_status_idle {
 
 sub update_child_status_busy {
   my ($self) = @_;
-
-# if (rand 4 < 1) { $self->report_backchannel_socket("I".pack("N",$self->{pid})."\n");return; warn "TEST for bug 4594"; die; }
   # "B  b1 b2 b3 b4 \n "
   $self->report_backchannel_socket("B".pack("N",$self->{pid})."\n");
 }
@@ -501,7 +545,7 @@ sub report_backchannel_socket {
   my ($self, $str) = @_;
   my $sock = $self->{backchannel}->get_parent_socket();
   $self->syswrite_with_retry($sock, $str, 'parent')
-        or write "syswrite() to parent failed: $!";
+        or die "syswrite() to parent failed: $!";
 }
 
 sub wait_for_orders {
@@ -627,14 +671,23 @@ retry_write:
       return undef;
     }
     else {
-      # give it 1 second to recover.  we retry indefinitely.
+      # give it 1 second to recover
       my $rout = '';
       vec($rout, $sock->fileno, 1) = 1;
       select(undef, $rout, undef, 1);
     }
   }
 
-  my $nbytes = $sock->syswrite($buf);
+  my $nbytes;
+  if (SUPPORT_TEST_INSTRUMENTATION && $TEST_MODE_CAUSE_RANDOM_WRITE_RETRIES &&
+            rand $TEST_MODE_CAUSE_RANDOM_WRITE_RETRIES < 1)
+  {
+    warn "prefork: TEST_MODE_CAUSE_RANDOM_WRITE_RETRIES simulating write failure";
+    $nbytes = undef; $! = &Errno::EAGAIN;
+  }
+  else {
+    $nbytes = $sock->syswrite($buf);
+  }
 
   if (!defined $nbytes) {
     unless ((exists &Errno::EAGAIN && $! == &Errno::EAGAIN)
@@ -686,6 +739,14 @@ sub adapt_num_children {
 
   foreach my $pid (@pids) {
     my $k = $kids->{$pid};
+
+    # note: race condition here.  if a child exits between the keys() call
+    # above, and this point, then $k will be undef here due to its deletion
+    # from the hash in the SIGCHLD handler.  This is harmless, but ugly, since
+    # it produces a 'Use of uninitialized value in numeric eq (==)' warning at
+    # the "== PFSTATE_IDLE" line below.
+    next unless defined $k;
+
     if ($k == PFSTATE_IDLE) {
       $statestr .= 'I';
       $num_idle++;
@@ -761,6 +822,20 @@ sub need_to_del_server {
   dbg("prefork: adjust: decreasing, too many idle children ($num_idle > $self->{max_idle}), killed $pid");
 }
 
+sub vec_all {
+  my ($self, $bitsref, $fhs, $value) = @_;
+  my $ret = 0;
+  foreach my $fh (@{$fhs}) {
+    next unless defined $fh;
+    if (defined $value) {
+      vec($$bitsref, $fh, 1) = $value;
+    } else {
+      $ret |= vec($$bitsref, $fh, 1);
+    }
+  }
+  return $ret;
+}
+
 1;
 
 __END__
