@@ -30,6 +30,12 @@ BEGIN {
 sub sa_t_init {
   my $tname = shift;
 
+  # optimisation -- don't setup spamd test parameters unless we're
+  # called "spamd_something" or "spamc_foo"
+  if ($tname !~ /spam[cd]/) {
+    $NO_SPAMD_REQUIRED = 1;
+  }
+
   if ($config{PERL_PATH}) {
     $perl_path = $config{PERL_PATH};
   }
@@ -46,18 +52,18 @@ sub sa_t_init {
   $perl_cmd .= " -w" if !defined($ENV{'TEST_PERL_WARN'})  or $ENV{'TEST_PERL_WARN'}  ne 'no';
 
   $scr = $ENV{'SCRIPT'};
-  $scr ||= "$perl_cmd ../spamassassin";
+  $scr ||= "$perl_cmd ../spamassassin.raw";
 
-  $spamd = "$perl_cmd ../spamd/spamd";
+  $spamd = "$perl_cmd ../spamd/spamd.raw";
 
   $spamc = $ENV{'SPAMC_SCRIPT'};
   $spamc ||= "../spamc/spamc";
 
   $salearn = $ENV{'SALEARN_SCRIPT'};
-  $salearn ||= "$perl_cmd ../sa-learn";
+  $salearn ||= "$perl_cmd ../sa-learn.raw";
 
   $spamdhost = $ENV{'SPAMD_HOST'};
-  $spamdhost ||= "localhost";
+  $spamdhost ||= "127.0.0.1";
   $spamdport = $ENV{'SPAMD_PORT'};
   $spamdport ||= probably_unused_spamd_port();
 
@@ -82,15 +88,17 @@ sub sa_t_init {
 
   read_config();
 
-  $NO_SPAMC_EXE = ($RUNNING_ON_WINDOWS &&
+  if (!$NO_SPAMD_REQUIRED) {
+    $NO_SPAMC_EXE = ($RUNNING_ON_WINDOWS &&
                    !$ENV{'SPAMC_SCRIPT'} &&
                    !(-e "../spamc/spamc.exe"));
-  $SKIP_SPAMC_TESTS = ($NO_SPAMC_EXE ||
+    $SKIP_SPAMC_TESTS = ($NO_SPAMC_EXE ||
                        ($RUNNING_ON_WINDOWS && !$ENV{'SPAMD_HOST'})); 
-  $SSL_AVAILABLE = ((!$SKIP_SPAMC_TESTS) &&  # no SSL test if no spamc
+    $SSL_AVAILABLE = ((!$SKIP_SPAMC_TESTS) &&  # no SSL test if no spamc
                     (!$SKIP_SPAMD_TESTS) &&  # or if no local spamd
                     (`$spamc -V` =~ /with SSL support/) &&
                     (`$spamd --version` =~ /with SSL support/));
+  }
   # do not remove prior test results!
   # rmtree ("log");
 
@@ -102,7 +110,7 @@ sub sa_t_init {
   rmtree ("log/test_rules_copy");
   mkdir ("log/test_rules_copy", 0755);
 
-  for $file (<../rules/*.cf>) {
+  for $file (<../rules/*.cf>, <../rules/*.pm>, <../rules/*.pre>) {
     $base = basename $file;
     copy ($file, "log/test_rules_copy/$base")
       or warn "cannot copy $file to log/test_rules_copy/$base";
@@ -138,13 +146,13 @@ sub sa_t_init {
 
   $ENV{'TEST_DIR'} = $cwd;
   $testname = $tname;
-
-  $current_user = (getpwuid($>))[0];
 }
 
 # a port number between 32768 and 65535; used to allow multiple test
 # suite runs on the same machine simultaneously
 sub probably_unused_spamd_port {
+  return 0 if $NO_SPAMD_REQUIRED;
+
   my $port;
   my @nstat = ();
   if (open(NSTAT, "netstat -a -n 2>&1 |")) {
@@ -234,8 +242,7 @@ sub sarun {
   rmtree ("log/outputdir.tmp"); # some tests use this
   mkdir ("log/outputdir.tmp", 0755);
 
-  %found = ();
-  %found_anti = ();
+  clear_pattern_counters();
 
   if (defined $ENV{'SA_ARGS'}) {
     $args = $ENV{'SA_ARGS'} . " ". $args;
@@ -246,10 +253,11 @@ sub sarun {
   my $scrargs = "$scr $args";
   $scrargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
   print ("\t$scrargs\n");
-  system ("$scrargs > log/$testname.${Test::ntest} $post_redir");
+  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  system ("$scrargs > log/d.$testname/${Test::ntest} $post_redir");
   $sa_exitcode = ($?>>8);
   if ($sa_exitcode != 0) { return undef; }
-  &checkfile ("$testname.${Test::ntest}", $read_sub) if (defined $read_sub);
+  &checkfile ("d.$testname/${Test::ntest}", $read_sub) if (defined $read_sub);
   1;
 }
 
@@ -280,10 +288,11 @@ sub salearnrun {
   my $salearnargs = "$salearn $args";
   $salearnargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
   print ("\t$salearnargs\n");
-  system ("$salearnargs > log/$testname.${Test::ntest}");
+  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  system ("$salearnargs > log/d.$testname/${Test::ntest}");
   $salearn_exitcode = ($?>>8);
   if ($salearn_exitcode != 0) { return undef; }
-  &checkfile ("$testname.${Test::ntest}", $read_sub) if (defined $read_sub);
+  &checkfile ("d.$testname/${Test::ntest}", $read_sub) if (defined $read_sub);
   1;
 }
 
@@ -319,10 +328,11 @@ sub spamcrun {
   $spamcargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
 
   print ("\t$spamcargs\n");
+  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
   if ($capture_stderr) {
-    system ("$spamcargs > log/$testname.out 2>&1");
+    system ("$spamcargs > log/d.$testname/out.${Test::ntest} 2>&1");
   } else {
-    system ("$spamcargs > log/$testname.out");
+    system ("$spamcargs > log/d.$testname/out.${Test::ntest}");
   }
 
   $sa_exitcode = ($?>>8);
@@ -330,7 +340,7 @@ sub spamcrun {
 
   %found = ();
   %found_anti = ();
-  &checkfile ("$testname.out", $read_sub) if (defined $read_sub);
+  &checkfile ("d.$testname/out.${Test::ntest}", $read_sub) if (defined $read_sub);
 
   ($sa_exitcode == 0);
 }
@@ -355,7 +365,8 @@ sub spamcrun_background {
   $spamcargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
 
   print ("\t$spamcargs &\n");
-  system ("$spamcargs > log/$testname.bg &") and return 0;
+  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  system ("$spamcargs > log/d.$testname/bg.${Test::ntest} &") and return 0;
 
   1;
 }
@@ -372,8 +383,9 @@ sub sdrun {
   1;
 }
 
+# out: $spamd_stderr
 sub start_spamd {
-
+  die "NO_SPAMD_REQUIRED in start_spamd! oops" if $NO_SPAMD_REQUIRED;
   return if $SKIP_SPAMD_TESTS;
 
   my $spamd_extra_args = shift;
@@ -410,14 +422,21 @@ sub start_spamd {
       qq{-p}, $spamdport,
     );
   }
+  if ($spamd_extra_args !~ /(?:--socketpath)/) {
+    push(@spamd_args,
+      qq{-A}, $spamdhost,
+    );
+  }
 
   if ($set_test_prefs) {
     warn "oops! SATest.pm: a test prefs file was created, but spamd isn't reading it\n";
   }
 
-  my $spamd_stdout = "log/$testname-spamd.out";
-  my $spamd_stderr = "log/$testname-spamd.err";
-  my $spamd_stdlog = "log/$testname-spamd.log";
+  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  my $spamd_stdout = "log/d.$testname/spamd.out.${Test::ntest}";
+     $spamd_stderr = "log/d.$testname/spamd.err.${Test::ntest}";    #  global
+  my $spamd_stdlog = "log/d.$testname/spamd.log.${Test::ntest}";
+
   my $spamd_forker = $ENV{'SPAMD_FORKER'}   ?
                        $ENV{'SPAMD_FORKER'} :
                      $RUNNING_ON_WINDOWS    ?
@@ -427,11 +446,12 @@ sub start_spamd {
                        $spamd_forker,
                        qq{SATest.pl},
                        qq{-Mredirect},
-                       qq{-o${spamd_stdout}},
                        qq{-O${spamd_stderr}},
+                       qq{-o${spamd_stdout}},
                        qq{--},
                        @spamd_args,
                        $spamd_extra_args,
+                       qq{-s ${spamd_stderr}.timestamped},
                        qq{&},
                     );
   unlink ($spamd_stdout, $spamd_stderr, $spamd_stdlog);
@@ -452,6 +472,12 @@ sub start_spamd {
         # Yes, DO retry on this error. I'm getting test failures otherwise
         # /Address already in use/ and $retries = 0;
 	/server pid: (\d+)/ and $spamd_pid = $1;
+
+        if (/ERROR/) {
+          warn "spamd error! $_";
+          $retries = 0; last;
+        }
+
 	$spamdlog .= $_;
       }
       close IN;
@@ -470,6 +496,7 @@ sub start_spamd {
 }
 
 sub stop_spamd {
+  die "NO_SPAMD_REQUIRED in stop_spamd! oops" if $NO_SPAMD_REQUIRED;
   return 0 if ( defined($spamd_already_killed) || $SKIP_SPAMD_TESTS);
 
   $spamd_pid ||= 0;
@@ -529,7 +556,14 @@ sub checkfile {
 
   # print "Checking $filename\n";
   if (!open (IN, "< log/$filename")) {
-    warn "cannot open log/$filename"; return undef;
+    # could be it already contains the "log/" prefix?
+    if (!open (IN, "< $filename")) {
+      warn "cannot open log/$filename or $filename"; return undef;
+    } else {
+      push @files_checked, "$filename";
+    }
+  } else {
+    push @files_checked, "log/$filename";
   }
   &$read_sub();
   close IN;
@@ -584,14 +618,22 @@ sub patterns_run_cb {
 }
 
 sub ok_all_patterns {
+  my ($dont_ok) = shift;
+
+  my $wasfailure = 0;
   foreach my $pat (sort keys %patterns) {
     my $type = $patterns{$pat};
     print "\tChecking $type\n";
     if (defined $found{$type}) {
-      ok ($found{$type} == 1) or warn "Found more than once: $type\n";
+      if (!$dont_ok) {
+        ok ($found{$type} == 1) or warn "Found more than once: $type\n";
+      }
     } else {
       warn "\tNot found: $type = $pat\n";
-      ok (0);                     # keep the right # of tests
+      if (!$dont_ok) {
+        ok (0);                     # keep the right # of tests
+      }
+      $wasfailure++;
     }
   }
   foreach my $pat (sort keys %anti_patterns) {
@@ -599,13 +641,21 @@ sub ok_all_patterns {
     print "\tChecking for anti-pattern $type\n";
     if (defined $found_anti{$type}) {
       warn "\tFound anti-pattern: $type = $pat\n";
-      ok (0);
+      if (!$dont_ok) { ok (0); }
+      $wasfailure++;
     }
     else
     {
-      ok (1);
+      if (!$dont_ok) { ok (1); }
     }
   }
+
+  if ($wasfailure) {
+    warn "Output can be examined in: ".join(' ', @files_checked)."\n";
+    return 0;
+  } else {
+    return 1;
+  }
 }
 
 sub skip_all_patterns {
@@ -642,6 +692,7 @@ sub skip_all_patterns {
 sub clear_pattern_counters {
   %found = ();
   %found_anti = ();
+  @files_checked = ();
 }
 
 sub read_config {
@@ -717,4 +768,72 @@ sub cleanup_safe_tmpdir {
   }
 }
 
+sub wait_for_file_to_change_or_disappear {
+  my ($f, $timeout, $action) = @_;
+
+  my $lastmod = (-M $f);
+
+  $action->();
+
+  my $wait = 0;
+  my $newlastmod;
+  do {
+    sleep (int($wait++ / 4) + 1) if $timeout > 0;
+    $timeout--;
+    $newlastmod = (-M $f);
+  } while((-e $f) && defined($newlastmod) &&
+                $newlastmod == $lastmod && $timeout);
+}
+
+sub wait_for_file_to_appear {
+  my ($f, $timeout) = @_;
+
+  # note that the wait period increases the longer it takes,
+  # 20 retries works out to a total of 60 seconds
+  my $wait = 0;
+  do {
+    sleep (int($wait++ / 4) + 1) if $timeout > 0;
+    $timeout--;
+  } while((!-e $f || -z $f) && $timeout);
+}
+
+sub read_from_pidfile {
+  my $f = shift;
+  my $npid = 0;
+  my $retries = 5;
+
+  do {
+    if ($retries != 5) {
+      sleep 1;
+      warn "retrying read of pidfile $f, due to short/nonexistent read: ".
+            "retry $retries";
+    }
+    $retries--;
+
+    if (!open (PID, "<".$f)) {
+      warn "Could not open pid file ${f}: $!\n";     # and retry
+      next;
+    }
+
+    $npid = <PID>;
+    if (defined $npid) { chomp $npid; }
+    close(PID);
+
+    if (!$npid || $npid < 1) {
+      warn "failed to read anything sensible from $f, retrying read";
+      $npid = 0;
+      next;
+    }
+    if (!kill (0, $npid)) {
+      warn "failed to kill -0 $npid, retrying read";
+      $npid = 0;
+    }
+
+  } until ($npid > 1 or $retries == 0);
+
+  return $npid;
+}
+
+sub dbgprint { print STDOUT "[".time()."] ".$_[0]; }
+
 1;
